using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Threading;
using System.Net;

namespace StradusUSBHID
{

  public class Util
  {

   

    public static readonly DateTime UnixEpochLocal = TimeZone.CurrentTimeZone.ToLocalTime(new DateTime(1970, 1, 1, 0, 0, 0));
    public static readonly DateTime UnixEpochUTC = TimeZone.CurrentTimeZone.ToUniversalTime(UnixEpochLocal);

    public static string nl = Environment.NewLine;

    #region Public Static Methods

    public static byte[] hex2bytes(string hex)
    {
      StringBuilder sb = new StringBuilder(hex.ToUpper());

      if (sb[0].Equals('0') && sb[1].Equals('X'))
      {
        sb.Remove(0, 2);
      }

      if (sb.Length % 2 != 0)
      {
        throw new ArgumentException("Invalid Hex String");
      }

      byte[] hexBytes = new byte[sb.Length / 2];
      try
      {
        for (int i = 0; i < hexBytes.Length; i++)
        {
          int stringIndex = i * 2;
          hexBytes[i] = Convert.ToByte(sb.ToString(stringIndex, 2), 16);
        }
      }
      catch (FormatException ex)
      {
        throw new ArgumentException("Invalid Hex String", ex);
      }

      return hexBytes;
    }

    public static string bytes2hex(byte[] bytes)
    {
      return bytes2hex(bytes, 0, bytes.Length);
    }

    public static string bytes2hex(byte[] bytes, bool prepend0x)
    {
      return bytes2hex(bytes, 0, bytes.Length, prepend0x);
    }

    public static string bytes2hex(byte[] bytes, int offset, int length)
    {
      return bytes2hex(bytes, offset, length, false);
    }

    public static string bytes2hex(byte[] bytes, int offset, int length, bool prepend0x)
    {
      if (bytes == null)
        return string.Empty;

      StringBuilder str = new StringBuilder();
      for (int i = offset; i < length; i++)
      {
        if (prepend0x)
          str.Append(" 0x");
        str.Append(bytes[i].ToString("X2"));
      }
      return str.ToString();
    }

    /// <summary>
    /// <para>Combines two byte arrays into one.</para>
    /// </summary>
    /// <param name="buffer1"><para>The prefixed bytes.</para></param>
    /// <param name="buffer2"><para>The suffixed bytes.</para></param>
    /// <returns><para>The combined byte arrays.</para></returns>
    public static byte[] CombineBytes(byte[] buffer1, byte[] buffer2)
    {
      byte[] combinedBytes = new byte[buffer1.Length + buffer2.Length];
      Buffer.BlockCopy(buffer1, 0, combinedBytes, 0, buffer1.Length);
      Buffer.BlockCopy(buffer2, 0, combinedBytes, buffer1.Length, buffer2.Length);

      return combinedBytes;
    }

    public static bool BytesAreSame(byte[] buf1, byte[] buf2)
    {
      if(buf1.Length != buf2.Length)
        return false;

      for (int i = 0; i < buf1.Length; i++)
      {
        if (buf1[i] != buf2[i])
          return false;
      }
      return true;
    }

    public static byte[] ReverseBytes(byte[] bytes)
    {
      if (bytes == null)
        return null;

      Array.Reverse(bytes);
      return bytes;
    }

    public static byte[] string2bytes(string str)
    {
      return Encoding.Default.GetBytes(str);
    }

    public static string bytes2string(byte[] bytes)
    {
      return Encoding.Default.GetString(bytes);
    }

    public static string string2hex(string str)
    {
      StringBuilder sb = new StringBuilder();
      char[] strs = str.ToCharArray();
      foreach (char chr in strs)
      {
        sb.Append(((byte)chr).ToString("X2"));
      }
      return sb.ToString();
    }

    public static UInt16 calcCRC(UInt16 crc, byte[] buffer, int index, int length)
    {
      byte b;

      for (int i = 0; i < length; i++)
      {
        b = buffer[i + index];
        crc = calcCRC(crc, b);
      }
      return crc;
    }

    public static UInt16 calcCRC(UInt16 crc, byte data)
    {
      UInt16 q;
      q = (UInt16)((crc ^ data) & 0xF);
      crc = (UInt16)((UInt16)(crc >> 4) ^ (q * 0x1081));
      q = (UInt16)((crc ^ (data >> 4)) & 0xF);
      crc = (UInt16)((UInt16)(crc >> 4) ^ (q * 0x1081));
      return crc;
    }

    /// <summary>
    /// converts local DateTime to UTC unix timestamp.
    /// </summary>
    /// <param name="time">local DateTime</param>
    /// <returns>UTC unix timestamp</returns>
    public static UInt32 TimeToUnix(DateTime time)
    {
      return (UInt32)((TimeSpan)TimeZone.CurrentTimeZone.ToUniversalTime(time).Subtract(UnixEpochUTC)).TotalSeconds;
    }

    /// <summary>
    /// converts UTC unix timestamp to local DateTime
    /// </summary>
    /// <param name="xtime">UTC unix timestamp</param>
    /// <returns>local DateTime</returns>
    public static DateTime UnixToTime(UInt32 xtime)
    {
      return TimeZone.CurrentTimeZone.ToLocalTime(UnixEpochUTC.AddSeconds(xtime));
    }

    public static object bytes2object(byte[] bytes, Type t)
    {
      object value = null;

      if (bytes == null)
        return value;

      switch (t.Name)
      {
        case "Byte":
          value = bytes[0];
          break;
        case "Byte[]":
          value = bytes;
          break;
        case "Int16":
          value = BitConverter.ToInt16(bytes, 0);
          break;
        case "UInt16":
          value = BitConverter.ToUInt16(bytes, 0);
          break;
        case "Int32":
          value = BitConverter.ToInt32(bytes, 0);
          break;
        case "UInt32":
          value = BitConverter.ToUInt32(bytes, 0);
          break;
        case "Int64":
          value = BitConverter.ToInt64(bytes, 0);
          break;
        case "UInt64":
          value = BitConverter.ToUInt64(bytes, 0);
          break;
        case "String":
          //value = Encoding.ASCII.GetString(bytes, 0, bytes.Length);
          value = Encoding.ASCII.GetString(bytes);
          break;
        case "Boolean":
          //if (bytes.Length == 0)
          //  value = false;
          //else
            value = BitConverter.ToBoolean(bytes, 0);
          break;
        case "DateTime":
          value = Util.UnixToTime(BitConverter.ToUInt32(bytes, 0));
          break;
        case "TimeSpan":
          value = new TimeSpan(0, 0, (int)BitConverter.ToUInt32(bytes, 0));
          break;
        case "IPAddress":
          Array.Reverse(bytes);
          value = new System.Net.IPAddress(bytes);
          break;
        case "Nullable":
        default:
          //throw new ApplicationException("Invalid Type data conversion");
          value = null;
          break;
      }

      return value;
    }

    public static byte[] object2bytes(object value)
    {
      byte[] bytes;

      if (value == null)
      {
        //string warn = "WARNING: converting NULL object value to byte[0]";
        //log.Warn(warn);
        //App.outs("PARSER: " + warn);
        return new byte[0];
      }

      Type t = value.GetType();

      switch (t.Name)
      {
        case "Byte":
          bytes = new byte[] { (byte)value };
          break;
        case "Byte[]":
          bytes = (byte[])value;
          break;
        case "Int16":
          bytes = BitConverter.GetBytes((Int16)value);
          break;
        case "UInt16":
          bytes = BitConverter.GetBytes((UInt16)value);
          break;
        case "Int32":
          bytes = BitConverter.GetBytes((Int32)value);
          break;
        case "UInt32":
          bytes = BitConverter.GetBytes((UInt32)value);
          break;
        case "Int64":
          bytes = BitConverter.GetBytes((Int64)value);
          break;
        case "UInt64":
          bytes = BitConverter.GetBytes((UInt64)value);
          break;
        case "String":
          if (((string)value).ToUpper().StartsWith("0X"))
            bytes = Util.hex2bytes((string)value);
          else
            bytes = Encoding.Default.GetBytes((String)value);
          break;
        case "Boolean":
          bytes = BitConverter.GetBytes((Boolean)value);
          break;
        case "DateTime":
          bytes = BitConverter.GetBytes(Util.TimeToUnix((DateTime)value));
          break;
        case "TimeSpan":
          bytes = BitConverter.GetBytes((UInt32)((TimeSpan)value).TotalSeconds);
          break;
        case "IPAddress":
          bytes = (value as System.Net.IPAddress).GetAddressBytes();
          Array.Reverse(bytes); 
          break;
        case "Nullable":
        default:
          //throw new ApplicationException("Invalid Type data conversion");
          bytes = new byte[0];
          break;
      }

      return bytes;
    }

    public static object object2object(object value, Type toType)
    {
      string fromType = value.GetType().Name;

      switch (toType.Name)
      {
        case "Int16":
          if (fromType == "UInt16")
            return (Int16)(UInt16)value;
          break;
        case "UInt16":
          if (fromType == "Int16")
            return (UInt16)(Int16)value;
          break;
        case "Int32":
          if (fromType == "UInt32")
            return (Int32)(UInt32)value;
          break;
        case "UInt32":
          if (fromType == "Int32")
            return (UInt32)(Int32)value;
          break;
        case "Int64":
          if (fromType == "UInt64")
            return (Int64)(UInt64)value;
          break;
        case "UInt64":
          if (fromType == "Int64")
            return (UInt64)(Int64)value;
          break;
        case "String":
          if (fromType == "IPAddress")
            return value.ToString();
          if (fromType == "Byte[]")
            return Util.bytes2hex(value as byte[]);
          break;
        case "IPAddress":
          if (fromType == "String")
            return IPAddress.Parse(value as string);
          break;
        case "Byte[]":
          if (fromType == "String")
            return Util.hex2bytes(value as string);
          break;
      }
      return value;
    }

    public static string object2string(object value)
    {
      string s;

      if (value == null)
        return "NULL";

      Type t = value.GetType();

      switch (t.Name)
      {
        case "Byte[]":
          s = Util.bytes2hex((byte[])value);
          break;
        case "DateTime":
          s = ((DateTime)value).ToUniversalTime().ToString("yyyy/MM/dd HH:mm:ss UTC");
          break;
        case "Nullable":
          s = "NULL";
          break;
        default:
          s = value.ToString();
          break;
      }

      return s;
    }

    #endregion



  }
}
